@@ -36,6 +36,8 @@
   #:use-module (shepherd system)
   #:use-module (shepherd args)
   #:use-module (shepherd comm)
+  #:autoload   (shepherd logger) (spawn-service-file-logger
+                                  spawn-service-system-logger)
   #:autoload   (ice-9 binary-ports) (put-u8 get-u8)
   #:autoload   (fibers operations) (choice-operation
                                     perform-operation
@@ -56,7 +58,7 @@ socket file at FILE-NAME upon exit of PROC.  Return the values of PROC."
                 (lambda args
                   (match args
                     ((key proc . _)
-                     (report-error (l10n "while opening socket '~a': ~a: ~a~%")
+                     (report-error (l10n "while opening socket '~a': ~a: ~a")
                                    file-name proc
                                    (strerror (system-error-errno args)))
                      ;; Stop services that were started from the config file
@@ -155,14 +157,6 @@ using fallback mechanism."))
            (fcntl fd F_SETFD (logior FD_CLOEXEC flags)))))
       (loop (+ fd 1)))))
 
-(cond-expand
- ((not guile-3)
-  (define exception-with-kind-and-args?
-    (const #f)))
- (else
-  (define exception-with-kind-and-args?
-    (exception-predicate &exception-with-kind-and-args))))
-
 (define (configuration-file-loader file)
   "Return a thunk that loads @var{file}, the user's configuration file."
   (define (failure)
@@ -172,35 +166,40 @@ using fallback mechanism."))
     #f)
 
   (define (handle-key-and-args-exception key args)
-    (match key
-      ('system-error
-       (local-output
-        (l10n "While loading configuration file '~a': ~a")
-        file (strerror (system-error-errno (cons key args)))))
-      (_
-       (local-output
-        (l10n "Uncaught exception while loading configuration file '~a': ~s")
-        file (cons key args)))))
+    (local-output
+     (l10n "While loading configuration file '~a': ~s")
+     file
+     (string-trim-right
+      (call-with-output-string
+        (lambda (port)
+          (print-exception port #f key args))))))
 
   (lambda ()
     (guard (c ((action-runtime-error? c)
                (local-output (l10n "action '~a' on service '~a' failed: ~s")
                              (action-runtime-error-action c)
-                             (action-runtime-error-service c)
+                             (service-canonical-name
+                              (action-runtime-error-service c))
                              (cons (action-runtime-error-key c)
                                    (action-runtime-error-arguments c)))
                (failure))
               ((exception-with-kind-and-args? c)
                (handle-key-and-args-exception
                 (exception-kind c) (exception-args c))
+               (failure))
+              ((message-condition? c)
+               (local-output
+                (l10n "Error while loading configuration file '~a': ~a")
+                file (condition-message c))
+               (failure))
+              (else
+               (local-output (l10n "Uncaught exception while loading \
+configuration file '~a': ~s")
+                             file c)
                (failure)))
-      (catch #t
-        (lambda ()
-          (load-in-user-module file)
-          (local-output (l10n "Configuration successfully loaded from '~a'.")
-                        file))
-        (lambda (key . args)                      ;for Guile 2.2
-          (handle-key-and-args-exception key args))))))
+      (load-in-user-module file)
+      (local-output (l10n "Configuration successfully loaded from '~a'.")
+                    file))))
 
 (define (get-u8/timeout port timeout)
   "Read a byte from @var{port}; return that byte, or the EOF object, or
@@ -315,6 +314,32 @@ using fallback mechanism."))
 
            (next-command))))))
 
+;; In Guile 3.0.x, 'call-with-input-file' & co. do not open their files as
+;; O_CLOEXEC.  The two procedures below address that.
+
+(define* (call-with-input-file/close-on-exec file proc
+                                             #:key
+                                             guess-encoding
+                                             encoding binary)
+  "Like @code{call-with-input-file}, but always open files as close-on-exec."
+  ;; Note: 'open-file' supports the "e" flag for O_CLOEXEC, but only since
+  ;; 3.0.9, hence the use of 'open'.
+  (call-with-port (open file (logior O_RDONLY O_CLOEXEC))
+    (lambda (port)
+      (cond (binary (set-port-encoding! port #f))
+            (encoding (set-port-encoding! port encoding))
+            (guess-encoding (set-port-encoding! port (file-encoding port))))
+      (proc port))))
+
+(define* (call-with-output-file/close-on-exec file proc
+                                              #:key encoding binary)
+  "Like @code{call-with-output-file}, but always open files as close-on-exec."
+  (call-with-port (open file (logior O_WRONLY O_CREAT O_CLOEXEC))
+    (lambda (port)
+      (cond (binary (set-port-encoding! port #f))
+            (encoding (set-port-encoding! port encoding)))
+      (proc port))))
+
 (define-syntax replace-core-bindings!
   (syntax-rules (<>)
     "Replace the given core bindings in the current process, restoring them upon
@@ -365,9 +390,16 @@ fork in the child process."
     ;; the signal thread.
     (maybe-signal-port %precious-signals))
 
-  (define log-flags
-    ;; Flags for 'open' when opening the log file.
-    (logior O_CREAT O_APPEND O_WRONLY O_CLOEXEC))
+  (define log-input+output
+    ;; Pipe used to send output to the logger of the 'root' service.
+    (match (pipe2 (logior O_NONBLOCK O_CLOEXEC))
+      ((input . output)
+       (let ((init! (lambda (port)
+                      (set-port-encoding! port "UTF-8")
+                      (set-port-conversion-strategy! port 'substitute))))
+         (init! input)
+         (init! output)
+         (cons input output)))))
 
   (initialize-cli)
 
@@ -375,6 +407,7 @@ fork in the child process."
 	(socket-file default-socket-file)
         (pid-file    #f)
         (secure      #t)
+        (silent?     #f)
         (logfile     #f))
     ;; Process command line arguments.
     (process-args (program-name) args
@@ -386,15 +419,13 @@ fork in the child process."
 		    #:takes-argument? #f
 		    #:description (l10n "synonym for --silent")
 		    #:action (lambda ()
-                               ;; XXX: Currently has no effect.
-                               #t))
+                               (set! silent? #t)))
 		  (option
 		    #:long-name "silent" #:short-name #\S
 		    #:takes-argument? #f
 		    #:description (l10n "don't do output to stdout")
 		    #:action (lambda ()
-                               ;; XXX: Currently has no effect.
-                               #t))
+                               (set! silent? #t)))
 		  (option
 		    ;; It might actually be desirable to have an
 		    ;; ``insecure'' setup in some circumstances, thus
@@ -439,25 +470,29 @@ fork in the child process."
     (and socket-file
          (verify-dir (dirname socket-file) #:secure? secure))
 
+    (define syslog?
+      ;; Is shepherd logging to /dev/log?
+      (and (not logfile) (= 1 (getpid))))
+
+    ;; By default, enable detailed deprecation warnings.
+    (unless (getenv "GUILE_WARN_DEPRECATED")
+      (debug-enable 'warn-deprecated))
+
     ;; Enable logging as first action.
-    (parameterize ((log-output-port
-                    (cond (logfile
-                           (buffering (open logfile log-flags)
-                                      'line))
-                          ((zero? (getuid))
-                           (syslog-output-port))
-                          (else
-                           (buffering (open (user-default-log-file) log-flags)
-                                      'line))))
+    (parameterize ((log-output-port (cdr log-input+output))
+
+                   ;; Send warnings such as deprecation warnings to the log.
+                   (current-warning-port (cdr log-input+output))
+
                    (%current-logfile-date-format
-                    (if (and (not logfile) (zero? (getuid)))
-                        (format #f "shepherd[~d]: " (getpid))
+                    (if syslog?
+                        ""                        ;for the "built-in" logger
                         default-logfile-date-format))
                    (%current-service-output-port
                     ;; Send output to log and clients.
                     (make-shepherd-output-port
-                     (if (and (zero? (getuid)) (not logfile))
-                         ;; By default we'd write both to /dev/kmsg and to
+                     (if (or silent? syslog?)
+                         ;; By default we'd write both to /dev/log and to
                          ;; stdout.  Redirect stdout to the bitbucket so we
                          ;; don't log twice.
                          (%make-void-port "w")
@@ -466,6 +501,11 @@ fork in the child process."
       (parameterize ((current-output-port (%current-service-output-port)))
         (set-port-encoding! (log-output-port) "UTF-8")
 
+        ;; Log provenance info.
+        (format #t "~a ~a (Guile ~a, ~a)~%"
+                package-name Version
+                (version) %host-type)
+
         (when (= 1 (getpid))
           ;; When running as PID 1, disable hard reboots upon ctrl-alt-del.
           ;; Instead, the kernel will send us SIGINT so that we can gracefully
@@ -499,6 +539,16 @@ fork in the child process."
                (register-services (list root-service))
                (start-service root-service)
 
+               (if syslog?
+                   (spawn-service-system-logger (car log-input+output)
+                                                #:service root-service)
+                   (spawn-service-file-logger (or logfile
+                                                  (if (= 1 (getpid))
+                                                      (system-default-log-file)
+                                                      (user-default-log-file)))
+                                              (car log-input+output)
+                                              #:service root-service))
+
                ;; Replace the default 'system*' binding with one that
                ;; cooperates instead of blocking on 'waitpid'.  Replace
                ;; 'primitive-load' (in C as of 3.0.9) with one that does
@@ -509,6 +559,8 @@ fork in the child process."
                            (spawn-command command)))
                 (system spawn-shell-command)
                 (primitive-load primitive-load*)
+                (call-with-input-file call-with-input-file/close-on-exec)
+                (call-with-output-file call-with-output-file/close-on-exec)
                 ((@ (guile) sleep) (@ (fibers) sleep)))
 
                (run-daemon #:socket-file socket-file
@@ -519,22 +571,6 @@ fork in the child process."
          #:parallelism 1                          ;don't create POSIX threads
          #:hz 0)))))       ;disable preemption, which would require POSIX threads
 
-(define (process-connection sock)
-  "Process client connection SOCK, reading and processing commands."
-  (catch 'system-error
-    (lambda ()
-      (match (read-command sock)
-        ((? shepherd-command? command)
-         (process-command command sock))
-        (#f                                    ;failed to read a valid command
-         #f))
-
-      ;; Currently we assume one command per connection.
-      (false-if-exception (close sock)))
-    (lambda args
-      ;; Maybe we got EPIPE while writing to SOCK, or something like that.
-      (false-if-exception (close sock)))))
-
 (define* (quit-exception-handler key #:optional value)
   "Handle the 'quit' exception, rebooting if we're running as root."
   ;; Note: The 'quit' exception does not necessarily have an associated value:
@@ -637,6 +673,22 @@ while evaluating @var{command}."
     (_
      (local-output (l10n "Invalid command.")))))
 
+(define (process-connection sock)
+  "Process client connection SOCK, reading and processing commands."
+  (catch 'system-error
+    (lambda ()
+      (match (read-command sock)
+        ((? shepherd-command? command)
+         (process-command command sock))
+        (#f                                    ;failed to read a valid command
+         #f))
+
+      ;; Currently we assume one command per connection.
+      (false-if-exception (close sock)))
+    (lambda args
+      ;; Maybe we got EPIPE while writing to SOCK, or something like that.
+      (false-if-exception (close sock)))))
+
 (define (process-textual-commands port)
   "Process textual commands from PORT.  'Textual' means that they're as you
 would write them on the 'herd' command line."
