@@ -22,13 +22,12 @@
 
 (define-module (shepherd support)
   #:use-module (shepherd config)
-  #:use-module (shepherd system)                 ;for 'O_CLOEXEC' on Guile 2.2
   #:autoload   (shepherd colors) (color-output? color colorize-string)
   #:use-module (ice-9 match)
   #:use-module (ice-9 format)
   #:autoload   (srfi srfi-1) (take)
   #:use-module (srfi srfi-9)
-  #:export (caught-error
+  #:export (exception-with-kind-and-args?
             assert
             let-loop
 
@@ -37,6 +36,7 @@
             ring-buffer-limit
             ring-buffer-insert
             ring-buffer->list
+            at-most
 
             buffering
             catch-system-error
@@ -44,6 +44,7 @@
             with-atomic-file-output
             mkdir-p
             with-directory-excursion
+            socket-address->string
 
             initialize-cli
 
@@ -53,12 +54,14 @@
             display-version
             program-name
             report-error
+            report-hint
             display-line
             non-blocking-port
             blocking-port
 
             user-homedir
             user-default-log-file
+            system-default-log-file
             default-logfile-date-format
             default-config-file
             default-socket-dir
@@ -75,20 +78,9 @@
 
             verify-dir))
 
-;; Report the caught error.
-;; FIXME: Needs some more work.
-(define (caught-error key args)
-  (case key
-    ((wrong-number-of-args)
-     (apply (lambda (subr fmt fmt-args data)
-	      (format #t "In ~a: " subr)
-	      (apply format #t fmt fmt-args)
-	      (newline))
-	    args))
-    (else
-     (display key)
-     (write args)
-     (newline))))
+(define exception-with-kind-and-args?
+  ;; This procedure is missing as of Guile 3.0.9.
+  (exception-predicate &exception-with-kind-and-args))
 
 (define-syntax assert
   (lambda (s)
@@ -271,6 +263,21 @@ given MODE."
      (lambda ()
        (chdir init)))))
 
+(define (socket-address->string address)
+  "Return a human-readable representation of ADDRESS, an object as returned by
+'make-socket-address'."
+  (let ((family (sockaddr:fam address)))
+    (cond ((= AF_INET family)
+           (string-append (inet-ntop AF_INET (sockaddr:addr address))
+                          ":" (number->string (sockaddr:port address))))
+          ((= AF_INET6 family)
+           (string-append "[" (inet-ntop AF_INET6 (sockaddr:addr address)) "]"
+                          ":" (number->string (sockaddr:port address))))
+          ((= AF_UNIX family)
+           (sockaddr:path address))
+          (else
+           (object->string address)))))
+
 
 
 (define %gettext-domain
@@ -316,26 +323,38 @@ There is NO WARRANTY, to the extent permitted by law.")))
   ;; Name of the program currently executing.
   (make-parameter "shepherd"))
 
-(define (print-error-message message)
-  "Print @var{message} to the current error port, prefixing it in standard GNU
+(define (message-printer color prefix)
+  (lambda (message)
+    "Print @var{message} to the current error port, prefixing it in standard GNU
 error format."
-  (define colorize
-    (if (color-output? (current-error-port))
-        (lambda (str)
-          (colorize-string str (color BOLD RED)))
-        identity))
+    (define colorize
+      (if (color-output? (current-error-port))
+          (lambda (str)
+            (colorize-string str color))
+          identity))
 
-  (format (current-error-port) "~a: ~a: ~a~%" (program-name)
-          (colorize (l10n "error")) message))
+    (format (current-error-port) "~a: ~a: ~a~%" (program-name)
+            (colorize (l10n prefix)) message)))
 
-(define-syntax report-error
-  (lambda (s)
-    "Report the given error message to stderr in standard GNU error format."
-    (syntax-case s (l10n)
-      ((_ (l10n message) args ...)
-       (string? (syntax->datum #'message))
+(define print-error-message
+  (let-syntax ((l10n (syntax-rules () ((_ str) str))))
+    (message-printer (color BOLD RED) (l10n "error"))))
 
-       #'(print-error-message (format #f (l10n message) args ...))))))
+(define print-hint
+  (let-syntax ((l10n (syntax-rules () ((_ str) str))))
+    (message-printer (color BOLD CYAN) (l10n "hint"))))
+
+(define-syntax-rule (define-reporter name printer)
+  (define-syntax name
+    (lambda (s)
+      "Report the given error message to stderr in standard GNU error format."
+      (syntax-case s (l10n)
+        ((_ (l10n message) args (... ...))
+         (string? (syntax->datum #'message))
+         #'(printer (format #f (l10n message) args (... ...))))))))
+
+(define-reporter report-error print-error-message)
+(define-reporter report-hint print-hint)
 
 (define (call-with-system-error-handling thunk)
   "Call THUNK, catching any 'system-error' exception."
@@ -433,6 +452,12 @@ TARGET should be a string representing a filepath + name."
   (mkdir-p %user-log-dir #o700)
   (string-append %user-log-dir "/shepherd.log"))
 
+(define (system-default-log-file)
+  "Return the file name of the log file of PID 1."
+  (let ((log-directory (in-vicinity %localstatedir "log")))
+    (mkdir-p log-directory #o700)
+    (in-vicinity log-directory "shepherd.log")))
+
 (define default-logfile-date-format
   ;; 'strftime' format string to prefix each entry in the log.
   "%Y-%m-%d %H:%M:%S ")
@@ -474,26 +499,18 @@ which has essential bindings pulled in."
   (let ((m (make-fresh-user-module)))
     ;; The typical configuration file wants to do '(service ...)', and
     ;; '(register-services ...)', so provide the relevant bindings by default.
-    (module-use! m (resolve-interface '(oop goops)))
     (module-use! m (resolve-interface '(shepherd service)))
     m))
 
-(cond-expand
- (guile-3 #t)
- (else
-  ;; This is missing in Guile 2.2.
-  (define (call-with-port port proc)
-    (call-with-values
-        (lambda () (proc port))
-      (lambda vals
-        (close-port port)
-        (apply values vals))))))
-
 (define (primitive-load* file)
   ;; Like 'primitive-load', but in Scheme, so that it does not introduce a
   ;; continuation barrier that would prevent code in FILE from suspending.
   (call-with-port (open file (logior O_CLOEXEC O_RDONLY))
     (lambda (port)
+      ;; Honor the "-*- coding: xyz -*-" cookie if any, read as UTF-8
+      ;; otherwise.
+      (set-port-encoding! port (or (file-encoding port) "UTF-8"))
+
       (let loop ((result *unspecified*))
         (match (read port)
           ((? eof-object?)
