@@ -59,7 +59,7 @@ $herd status transient-test1 && false
 
 # Terminate the service and make sure it gets unregistered.
 $herd status transient-test2 | grep "transient, running"
-kill $($herd status transient-test2 | grep Running | sed -e's/^.* \([0-9]\+\).*$/\1/g')
+kill $($herd status transient-test2 | grep PID | sed -e's/^.* \([0-9]\+\) running.*$/\1/g')
 $herd status transient-test2 && false
 
 test $($herd status | grep transient-test | wc -l) -eq 0
